#!/usr/bin/env python
# setup.py generated by flit for tools that don't yet use PEP 517

from distutils.core import setup

packages = \
['incant']

package_data = \
{'': ['*']}

package_dir = \
{'': 'src'}

extras_require = \
{'dev': ['black',
         'cattrs',
         'coverage',
         'flake8',
         'httpx',
         'isort',
         'mypy',
         'pytest-asyncio',
         'pytest',
         'quart',
         'quattro',
         'rich',
         'structlog',
         'uvicorn']}

setup(name='incant',
      version='0.3',
      description='Magical function composition',
      author=None,
      author_email='Tin Tvrtkovic <tinchester@gmail.com>',
      url=None,
      packages=packages,
      package_data=package_data,
      package_dir=package_dir,
      extras_require=extras_require,
      python_requires='>=3.8',
     )
